import _extends from '@babel/runtime/helpers/esm/extends'
import _objectWithoutProperties from '@babel/runtime/helpers/esm/objectWithoutProperties'
import React from 'react'
import {CDLL} from 'cdll-memoize'
import PropTypes from 'prop-types'
import {getChunkScripts, findChunks} from './utils' // context is necessary for keeping track of which components in the
// current react tree depend on which corresponding chunks/promises

const {Provider, Consumer} = React.createContext({})
export const WAITING = 0b0 // promise has not yet started loading

export const LOADING = 0b1 // promise has started loading

export const REJECTED = 0b10 // promise was rejected

export const RESOLVED = 0b11 // promise was successfully resolved
// tracks the chunks used in the rendering of a tree

export function createChunkCache() {
  let map = {}
  const cache = {
    get: k => map[k],
    set: (k, v) => (map[k] = v),
    invalidate: k => delete map[k],
    // returns an array of chunk names used by the current react tree
    getChunkNames: () => Object.keys(map),
    // returns a Set of Webpack chunk objects used by the current react tree
    getChunks: stats => findChunks(stats, cache.getChunkNames()),
    // returns a string of <script> tags for Webpack chunks used by the
    // current react tree
    getChunkScripts: (stats, opt = {}) => getChunkScripts(stats, cache, opt),
  }

  function _ref2(fn) {
    return Object.keys(map).forEach(k => fn(k, map[k]))
  }

  if (process.env.NODE_ENV !== 'production') {
    cache.forEach = _ref2
  }

  return cache
} // loads an array of Lazy components

function _ref3(i) {
  return i.load()
}

export function load(...instances) {
  return Promise.all(instances.map(_ref3))
} // preloads all of the async components used in the current react tree

export class WaitForPromises {
  constructor() {
    this.chunkPromises = []
  }

  load() {
    return Promise.all(this.chunkPromises).then(() => (this.chunkPromises = []))
  }
}
export function loadAll(
  app,
  render = require('react-dom/server').renderToStaticMarkup
) {
  const waitForPromises = new WaitForPromises()

  class WaitForPromisesProvider extends React.Component {
    getChildContext() {
      return {
        waitForPromises,
      }
    }

    render() {
      return app
    }
  }

  WaitForPromisesProvider.childContextTypes = {
    waitForPromises: PropTypes.object,
  }

  var _ref =
    /*#__PURE__*/
    React.createElement(WaitForPromisesProvider, null)

  function process() {
    const html = render(_ref)
    return waitForPromises.chunkPromises.length > 0
      ? waitForPromises.load().then(process)
      : html
  }

  return Promise.resolve().then(process)
} // the purpose of this function is to avoid a flash or loading
// spinner when your app initially hydrates/renders

export function loadInitial(chunkCache = globalChunkCache) {
  let chunks = document.getElementById('__INITIAL_BROKER_CHUNKS__')

  if (!chunks) {
    throw new Error(
      'No chunk cache element was found at <script id="__INITIAL_BROKER_CHUNKS__">'
    )
  }

  chunks = JSON.parse(chunks.firstChild.data)
  const loading = [] // preloads the chunk scripts

  for (let script of document.querySelectorAll('script[data-rb]')) {
    loading.push(
      new Promise(resolve => {
        if (script.getAttribute('data-loaded')) {
          resolve()
        } else {
          script.addEventListener('load', resolve)
        }
      })
    )
  }

  function _ref4(chunkName) {
    let component

    try {
      component = __webpack_require__(chunks[chunkName]).default
    } finally {
      // sets the component in the chunk cache if it is valid
      if (typeof component === 'function') {
        if (typeof module !== 'undefined' && module.hot) {
          __webpack_require__.c[chunks[chunkName]].hot.accept()
        }

        chunkCache.set(chunkName, {
          status: RESOLVED,
          lazy: new CDLL([]),
          component,
        })
      }
    }
  }

  return Promise.all(loading).then(() => Object.keys(chunks).forEach(_ref4))
}
const globalChunkCache = createChunkCache()

function _ref8(chunkName, chunk) {
  if (chunk.status !== WAITING && chunk.status !== LOADING) {
    chunk.status = WAITING
    console.log(' -', chunkName)
  }
}

export class LazyProvider extends React.Component {
  constructor(props) {
    super(props) // if there wasn't a chunkCache explicitly provided, one is created

    this.getStatus = chunkName => {
      // gets the cached status of a given chunk name
      const chunk = this.chunkCache.get(chunkName)
      return chunk === void 0 ? WAITING : chunk.status
    }

    this.getComponent = chunkName => {
      // gets the cached component for a given chunk name
      const chunk = this.chunkCache.get(chunkName)
      return chunk && chunk.component
    }

    this.subscribe = (chunkName, lazyComponent) => {
      // subscribes a consumer to @chunkName and updates the consumer's state
      // when the chunk has resolved
      const chunk = this.chunkCache.get(chunkName)

      if (chunk === void 0 || chunk.status === WAITING) {
        // a circular doubly linked list is used for maintaining the consumers
        // listening to a chunk's resolution because there are far fewer
        // operations in deleting a consumer from the listeners than you'd
        // have with a plain Array
        const lazy = new CDLL([lazyComponent])
        this.chunkCache.set(chunkName, {
          status: WAITING,
          lazy,
        })
        const promise = lazyComponent.promise()

        if (this.context && this.context.waitForPromises) {
          this.context.waitForPromises.chunkPromises.push(promise)
        }

        return this.load(chunkName, promise)
      } else {
        // this chunk is already being listened to so all we need to do is add
        // the consumer to the list of consumers that need to be updated
        // once the chunk's promise resolves
        chunk.lazy.push(lazyComponent)
        return chunk.promise
      }
    }

    this.unsubscribe = (chunkName, lazyComponent) => {
      // removes a consumer from the list of consumers listening for a chunk's
      // resolution
      const chunk = this.chunkCache.get(chunkName)
      const element = chunk.lazy.find(lazyComponent)
      element !== void 0 && chunk.lazy.delete(element)
    }

    this.load = (chunkName, promise) => {
      // loads a given chunk and updates its consumers when it is resolved or
      // rejected. also sets the chunk's status to 'LOADING' if it hasn't
      // already resolved
      const chunk = this.chunkCache.get(chunkName)

      function _ref5(c) {
        return c.resolving(chunkName)
      }

      if (chunk.status === WAITING) {
        // tells subscribed components that we've started loading this chunk
        chunk.promise = promise
          .then(component => this.resolved(chunkName, component))
          .catch(err => this.rejected(chunkName, err))
        chunk.status = LOADING
        chunk.lazy.forEach(_ref5)
      }

      return chunk.promise
    }

    this.resolved = (chunkName, component) => {
      // updates a chunk's consumers when the chunk is resolved and sets the
      // chunk status to 'RESOLVED'
      const chunk = this.chunkCache.get(chunkName)

      function _ref6(c) {
        return c.resolved(chunkName, chunk.component)
      }

      if (chunk.status !== RESOLVED) {
        chunk.status = RESOLVED // modules typically resolve with a 'default' attribute, but some don't.
        // likewise, fetch() never resolves with a 'default' attribute.

        chunk.component =
          component && component.default !== void 0
            ? component.default
            : component // updates each chunk listener with the resolved component

        chunk.lazy.forEach(_ref6)
      }

      return chunk.component
    }

    this.rejected = (chunkName, err) => {
      // updates a chunk's consumers when the chunk is rejected and sets the
      // chunk status to 'REJECTED'
      const chunk = this.chunkCache.get(chunkName)

      function _ref7(c) {
        return c.rejected(chunkName, err)
      }

      if (chunk.status !== REJECTED) {
        chunk.status = REJECTED // updates each chunk listener with the caught error

        chunk.lazy.forEach(_ref7)
      }

      return err
    }

    this.chunkCache = props.chunkCache || globalChunkCache // this is the context that is provided to the LazyConsumers

    this.providerContext = {
      load: this.load,
      subscribe: this.subscribe,
      unsubscribe: this.unsubscribe,
      getStatus: this.getStatus,
      getComponent: this.getComponent,
    }
  }

  componentDidMount() {
    // this clears the chunk cache when HMR disposes of a module
    if (process.env.NODE_ENV !== 'production') {
      if (typeof module !== 'undefined' && module.hot) {
        this.invalidateChunks = status => {
          if (status === 'idle') {
            // fetches any preloaded chunks
            console.log('[Broker HMR] reloading')
            let chunks = document.getElementById('__INITIAL_BROKER_CHUNKS__')

            if (!!chunks) {
              // initial chunks were loaded and we need this workaround to get them to
              // refresh for some reason
              chunks = JSON.parse(chunks.firstChild.data)
              Object.keys(chunks).forEach(chunkName => {
                if (typeof module !== 'undefined' && module.hot) {
                  let component

                  try {
                    component = __webpack_require__(chunks[chunkName]).default
                  } finally {
                    const chunk = this.chunkCache.get(chunkName)
                    chunk.status = WAITING
                    this.load(
                      chunkName,
                      Promise.resolve(
                        __webpack_require__.c[chunks[chunkName]].exports
                      )
                    )

                    __webpack_require__.c[chunks[chunkName]].hot.accept()

                    console.log(' -', chunkName)
                  }
                }
              })
            }
          } else if (status === 'apply') {
            this.chunkCache.forEach(_ref8)
          }
        }

        module.hot.addStatusHandler(this.invalidateChunks)
      }
    }
  }

  componentWillUnmount() {
    if (process.env.NODE_ENV !== 'production') {
      typeof module !== 'undefined' &&
        module.hot &&
        module.hot.removeStatusHandler(this.invalidateChunks)
    }
  }

  render() {
    const children = React.Children.only(this.props.children)
    return React.createElement(Provider, {
      value: this.providerContext,
      children: children,
    })
  }
}
LazyProvider.contextTypes = {
  waitForPromises: PropTypes.object,
}
const defaultOpt = {
  loading: null,
  error: null,
}

function _ref9({status}) {
  return status === LOADING
    ? null
    : {
        status: LOADING,
        component: null,
        error: null,
      }
}

function _ref10(p) {
  return p()
}

export default function lazy(promises, opt = defaultOpt) {
  const chunkName = Object.keys(promises)[0]

  class Lazy extends React.Component {
    // since Promise isn't cancellable this is necessary for avoiding
    // 'update on unmounted component' errors in React
    // the promise function tied to this chunk
    // caches the name of the chunk
    constructor(props) {
      super(props) // subscribes the component to changes in the chunk's status

      this.mounted = false
      this.unmounted = false
      this.promise = promises[chunkName]
      this.chunkName = chunkName

      this.resolving = () =>
        this.unmounted === false &&
        this.mounted === true &&
        this.setState(_ref9)

      this.resolved = (chunkName, resolvedComponent) =>
        this.unmounted === false &&
        this.mounted === true &&
        this.setState({
          status: RESOLVED,
          component: resolvedComponent,
          error: null,
        })

      this.rejected = (chunkName, error) =>
        this.unmounted === false &&
        this.mounted === true &&
        this.setState({
          status: REJECTED,
          error,
        })

      this.load = () => this.props.lazy.load(this.chunkName, this)

      this.props.lazy.subscribe(this.chunkName, this) // gets the initial status of the chunk in checking whether or not
      // its already been subscribed/resolved

      const _status = this.props.lazy.getStatus(this.chunkName) // retrieves the initial component if there is one for this chunk

      const component = this.props.lazy.getComponent(this.chunkName) // sets the initial state

      this.state = {
        status: _status,
        component,
        error: null,
      }
    }

    componentDidMount() {
      this.mounted = true
      const status = this.props.lazy.getStatus(this.chunkName),
        component = this.props.lazy.getComponent(this.chunkName)
      if (this.state.status !== status || this.state.component !== component)
        this.setState({
          status,
          component: status === LOADING ? null : component,
          error: status !== REJECTED ? null : this.state.error,
        })
    }

    componentWillUnmount() {
      this.unmounted = true
      this.props.lazy.unsubscribe(this.chunkName, this)
    } // sets LOADING status for chunks that are currently resolving

    render() {
      let {status, component, error} = this.state

      let _this$props = this.props,
        {children, lazy} = _this$props,
        props = _objectWithoutProperties(_this$props, ['children', 'lazy'])

      switch (status) {
        case WAITING:
        case LOADING:
          // returns 'loading' component
          return opt.loading
            ? opt.loading(props, {
                retry: this.load,
                error: null,
              })
            : null

        case REJECTED:
          // returns 'error' component
          // if there isn't an explicitly defined 'error' component, the
          // 'loading' component will be used as a backup with the error
          // message passed in the second argument
          const render = opt.loading || opt.error
          return render
            ? render(props, {
                retry: this.load,
                error,
              })
            : null

        case RESOLVED:
          // returns the proper resolved component
          return React.createElement(component, props, children)
      }
    }
  } // wraps the Lazy component with the context consumer so that it gets updated
  // when chunk status changes

  function LazyConsumer(props) {
    return React.createElement(Consumer, {
      children: cxt =>
        React.createElement(
          Lazy,
          _extends(
            {
              lazy: cxt,
            },
            props
          )
        ),
    })
  } // necessary for calling Component.load from the application code

  LazyConsumer.load = () => Promise.all(Object.values(promises).map(_ref10)) // <Lazy(pages/Home)> makes visual grep'ing easier in react-dev-tools

  if (process.env.NODE_ENV !== 'production') {
    LazyConsumer.displayName = 'Lazy(' + Object.keys(promises).join(', ') + ')'
  }

  return LazyConsumer
} // the lazy object acts like 'exports' here

lazy.load = load
lazy.loadAll = loadAll
lazy.loadInitial = loadInitial
lazy.createChunkCache = createChunkCache
lazy.Provider = LazyProvider
lazy.WAITING = WAITING
lazy.LOADING = LOADING
lazy.REJECTED = REJECTED
lazy.RESOLVED = RESOLVED
